# -*- coding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    Copyright (C) 2013 credativ (http://www.credativ.co.uk). All Rights Reserved
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import re, time, datetime
from openerp.osv import osv
import pooler
from openerp.addons.base_external_referentials.external_osv import MappingError
from tools.translate import _
from pyactiveresource.activeresource import ActiveResource
from pyactiveresource.connection import ConnectionError
from psycopg2 import IntegrityError
from tools.misc import DEFAULT_SERVER_DATETIME_FORMAT

import logging
_logger = logging.getLogger(__name__)
_logger.setLevel(logging.DEBUG)

def clean_read_m2o(record):
    cleaned = {}
    for k, v in record.items():
        if isinstance(v, tuple):
            cleaned[k] = v[0]
        else:
            cleaned[k] = v
    return cleaned

def rm_m2o_get_id(record, field):
    '''
    ActiveResource one2many fields are returned as dicts containing
    the keys 'id' and (sometimes) 'name'. This function returns the
    'id'.
    '''
    if record.get(field, False) and isinstance(record[field], dict) and 'id' in record[field]:
        return int(record[field]['id'])
    else:
        return None

def rm_convert_m2o_fields(record):
    '''
    This function converts any one2many dicts in an ActiveResource
    record into int ids.
    '''
    res = record.copy()
    res.update(dict([(f, rm_m2o_get_id(record, f))
                     for f in record.keys()
                     if rm_m2o_get_id(record, f)]))
    return res

class Connection(object):
    '''
    This class is a wrapper around the ActiveResource REST API to
    Redmine. Its primary point of entry is the call method. This
    method's 'method' parameter should be given a string of the form
    "ModelName.method", e.g. "Project.find".
    '''

    def __init__(self, location, username, password, api_uri, api_key, DEBUG=False):
        if api_uri and api_key:
            self._site = api_uri
            self._api_key = api_key
        elif location and username and password:
            self._site = location
            self._user = username
            self._password = password
        else:
            raise osv.except_osv('Value Error',
                                 'Invalid Redmine details:\n\nURL: %s\nUser: %s\nPassword: %s\nAPI URI: %s\nAPI key: %s' %\
                                 (location, username, password, api_uri, api_key))

    def connect(self):
        '''
        The REST API is stateless, so there's no need to establish a
        connection. Therefore, this method just checks that the REST
        API is accessible.
        '''
        return self.is_accessible()

    def is_accessible(self):
        try:
            test_resource = self.get_model('Project')
            test_resource.find()
            return True
        except Exception, e:
            raise osv.except_osv(_('Connection error'),
                                 _('Could not connect to redmine at %s\n\n"%s"' % (test_resource._site, e.message)))

    def get_model(self, model_name):
        # create a new class of the given model_name and with
        # ActiveResource as its base class
        if hasattr(self, '_api_key'):
            access = {'_site': self._site, '_headers': {'X-Redmine-API-Key': self._api_key}}
        elif hasattr(self, '_user') and hasattr(self, '_password'):
            access = {'_site': self._site, '_user': self._user, '_password': self._password}
        else:
            raise osv.except_osv(_('Authentication error'),
                                 _('No applicable authentication method for Redmine API.'))
        model_class = type(str(model_name), (ActiveResource,), access)
        return model_class()
        
    def call(self, method, *args, **kwargs):
        '''
        Calls the specified method on the external system.

        @param method (str): The method argument must be prefixed
        with the external model name.
        '''
        m = re.search(r'([A-Za-z_][A-Za-z_0-9]*)\.([A-Za-z_][A-Za-z_0-9]*)', method)
        if not m:
            raise osv.except_osv(_('Value Error'),
                                 _('The method argument must be of the form "ModelName.method_name"; got "%s"' % (method,)))

        model_name, method_name = m.groups()
        method = getattr(self.get_model(model_name), method_name, False)
        if not method:
            raise osv.except_osv(_('Attribute Error'),
                                 _('The external model %s has no method "%s"' % (model_name, method_name)))

        # FIXME PyActiveResource does not provide access to the
        # metadata (such as 'type', 'total_count', 'limit', 'offset')
        # returned by Redmine. Redmine limits to 25 records by
        # default. Ideally we'd find the total_count then retrieve
        # with that as limit. Instead, we're going to set a high limit
        # and carry on retrieving until less than $limit records are
        # returned.
        if isinstance(args, (int, long)) or isinstance(args[0], (int, long)):
            return method(*args, **kwargs)
        else:
            limit = count = 1000
            offset = 0
            res = []
            while count == limit:
                kwargs['limit'] = limit
                kwargs['offset'] = offset
                next_recs = method(*args, **kwargs)
                res.extend(next_recs)
                count = len(next_recs)
                offset += count
            return res

class redmine_osv(osv.osv):
    '''
    This class is used as a mixin for models which need to be able to
    synchronise with Redmine.
    '''
    _register = False

    _RM_FIELD = 'redmine_id'

    def rm_to_oe(self, cr, uid, rm_id, referential_id, *arguments):
        pass

    def record_save(self, cr, uid, rec_id, vals, defaults):
        if defaults:
            for key in defaults.keys():
                vals[key] = defaults[key]
        if rec_id:
            self.write(cr, uid, rec_id[0], vals)
        else:
            self.create(cr, uid, vals)

    def sync_import(self, cr, uid, rm_records, referential_id, defaults=None, *attrs):
        pass

    def rm_import_base(self, cr, uid, conn, referential_id, defaults=None, context=None):
        if context is None:
            context = {}
        if defaults is None:
            defaults = {}
        if not 'ids_or_filter' in context.keys():
            context['ids_or_filter'] = []

        result = {'create_ids': [], 'write_ids': []}

        mapping_id = self.pool.get('external.mapping').search(cr, uid, [('model','=',self._name),
                                                                        ('referential_id','=',referential_id)],
                                                              context=context)
        if mapping_id:
            data = []
            if context.get('id', False):
                get_method = self.pool.get('external.mapping').read(cr, uid, mapping_id[0], ['external_get_method']).get('external_get_method', False)
                if get_method:
                    data = conn.call(get_method, context.get('id', False)).to_dict()
                    data['external_id'] = context.get('id', False)
                    result = self.ext_import(cr, uid, [data], referential_id, defaults, context)
            else:
                list_method = self.pool.get('external.mapping').read(cr, uid, mapping_id[0], ['external_list_method']).get('external_list_method', False)
                if list_method:
                    data = [rec.to_dict() for rec in conn.call(list_method, context['ids_or_filter'])]

                    if context.get('one_by_one', False):
                        self.rm_import_one_by_one(cr, uid, conn, referential_id, mapping_id[0], data, defaults, context)
                    else:
                        result = self.ext_import(cr, uid, data, referential_id, defaults, context)

        return result

    def rm_import_one(self, cr, uid, conn, record, referential_id, mapping_id, defaults=None, context=None):
        ext_id = record[self.pool.get('external.mapping').read(cr, uid, mapping_id, ['external_key_name'])['external_key_name']]
        get_method = self.pool.get('external.mapping').read(cr, uid, mapping_id, ['external_get_method']).get('external_get_method', False)
        rec_data = [conn.call(get_method, ext_id)]
        rec_result = self.ext_import(cr, uid, rec_data, referential_id, defaults, context)
        return rec_result['create_ids'], rec_result['write_ids']

    def rm_import_one_by_one(self, cr, uid, conn, referential_id, mapping_id, data, defaults=None, context=None):
        if context is None:
            context = {}

        report_line_obj = self.pool.get('external.report.line')
        result = {'create_ids': [], 'write_ids': []}
        import_ctx = context.copy()
        import_ctx.pop('one_by_one', False)
        import_ctx.pop('use_external_log', False)
        import_ctx['import_no_new_cr'] = True

        for record in data:
            record_cr = pooler.get_db(cr.dbname).cursor()
            try:
                ext_id = record[self.pool.get('external.mapping').read(
                    record_cr, uid, mapping_id, ['external_key_name'])['external_key_name']]
                cids, wids = self.rm_import_one(
                    record_cr, uid, conn, record, referential_id, mapping_id, defaults=defaults, context=import_ctx)
            except (MappingError, osv.except_osv, IntegrityError):
                record_cr.rollback()
                report_line_obj.log_failed(cr, uid, self._name, 'import',
                    referential_id, external_id=ext_id, defaults=defaults, context=context)
            else:
                record_cr.commit()
                report_line_obj.log_success(cr, uid, self._name, 'import',
                    referential_id, external_id=ext_id, context=context)
                result['create_ids'].append(cids)
                result['write_ids'].append(wids)
            finally:
                record_cr.close()
        return result

    def get_external_data(self, cr, uid, conn, external_referential_id, defaults=None, context=None):
        if context is None:
            context = {}
        context = dict(context)
        return self.rm_import_base(cr, uid, conn, external_referential_id, defaults, context)

    def ext_create(self, cr, uid, data, conn, method, oe_id, context):
        conn.call(method, data)

    def extid_to_oeid(self, cr, uid, rm_id, referential_id, context=None):
        if rm_id:
            existing_id = self.extid_to_existing_oeid(cr, uid, rm_id, referential_id, context=context)
            if existing_id:
                return existing_id
            try:
                conn = self.pool.get('external.referential').external_connection(cr, uid, referential_id)
                ctx = {'id':id}
                if context.get('import_no_new_cr', False):
                    ctx.update({'import_no_new_cr': True})
                result = self.get_external_data(cr, uid, conn , referential_id, {}, ctx)
                if len(result['create_ids']) == 1:
                    return result['create_ids'][0]
            except Exception, error:
                raise osv.except_osv(_('External synchronisation'),
                                     _('Error when importing on fly the object %s with the external_id %s and the external referential %s.\n Error : %s' %\
                                       (self._name, rm_id, referential_id, error)))
        return False

    def ext_export_one(self, cr, uid, record_data, referential_id, defaults=None, context=None):
        if context is None:
            context = {}
        if defaults is None:
            defaults = {}
    
        mapping_pool = self.pool.get('external.mapping')
        map_line_pool = self.pool.get('external.mapping.line')
        data_pool = self.pool.get('ir.model.data')
    
        create_id = write_id = False
        mapping_ids =\
            mapping_pool.search(cr, uid,
                                [('model','=',self._name),
                                 ('referential_id','=',referential_id)],
                                limit=1, context=context)
        if mapping_ids:
            mapping_id = mapping_ids[0]
            mapping_rec = mapping_pool.read(cr, uid, mapping_id,
                                            ['external_update_method', 'external_create_method'],
                                            context=context)

            mapping_line_ids = map_line_pool.search(cr, uid,
                                                    [('mapping_id','=',mapping_id),
                                                     ('type','in',['in_out', 'out'])],
                                                    context=context)
            mapping_lines = map_line_pool.read(cr, uid, mapping_line_ids,
                                               ['external_field', 'out_function'],
                                               context=context)
    
            if mapping_lines:
                ext_data = self.extdata_from_oevals(cr, uid, referential_id, record_data, mapping_lines, defaults, context=context)
    
                rec_check_ids = data_pool.search(cr, uid,
                                                 [('model','=',self._name),
                                                  ('res_id','=',record_data['id']),
                                                  ('module','ilike','extref'),
                                                  ('external_referential_id','=',referential_id)],
                                                 context=context)
    
                conn = context.get('conn_obj', False) or\
                       self.pool.get('external.referential').external_connection(cr, uid, referential_id)
                if mapping_rec and len(rec_check_ids) >= 1:
                    ext_id = self.oeid_to_extid(cr, uid, record_data['id'], referential_id, context=context)

                    if not context.get('force', False):
                        #Record exists, check if update is required, for that collect last update times from ir.data & record
                        last_exported_times = data_pool.read(cr, uid, rec_check_ids[0], ['write_date', 'create_date'], context=context)
                        last_exported_time = last_exported_times.get('write_date', False) or last_exported_times.get('create_date', False)
                        this_record_times = self.read(cr, uid, record_data['id'], ['write_date', 'create_date'], context=context)
                        last_updated_time = this_record_times.get('write_date', False) or this_record_times.get('create_date', False)

                        if not last_updated_time:
                            cr.execute("select write_date, create_date from %s where id=%s;" % (self._name.replace('.', '_'), record_data['id']))
                            read = cr.fetchone()
                            last_updated_time = read[0] and read[0].split('.')[0] or read[1] and read[1].split('.')[0] or False

                        if last_updated_time and last_exported_time:
                            last_exported_time = datetime.datetime.fromtimestamp(time.mktime(time.strptime(last_exported_time[:19], DEFAULT_SERVER_DATETIME_FORMAT)))
                            last_updated_time = datetime.datetime.fromtimestamp(time.mktime(time.strptime(last_updated_time[:19], DEFAULT_SERVER_DATETIME_FORMAT)))
                            if last_exported_time + datetime.timedelta(seconds=1) > last_updated_time:
                                # skip update for this resource
                                return False, False

                    if mapping_rec['external_update_method']:
                        try:
                            self.ext_update(cr, uid, ext_data, conn, mapping_rec['external_update_method'], record_data['id'],
                                            ext_id, rec_check_ids[0], mapping_rec['external_create_method'], context)
                            write_id = record_data['id']
                            ir_model_data_vals = {'res_id': record_data['id']}
                            data_pool.write(cr, uid, rec_check_ids[0], ir_model_data_vals)
                            _logger.info("Updated in External Ref %s from OpenERP with external_id %s and OpenERP id %s successfully",
                                         self._name, ext_id, record_data['id'])
                        except Exception, err:
                            _logger.error("Failed to Update in External Ref %s from OpenERP with external_id %s and OpenERP id %s",
                                          self._name, ext_id, record_data['id'])
                            raise
                else:
                    # Record needs to be created
                    if mapping_rec['external_create_method']:
                        try:
                            crid = self.ext_create(cr, uid, ext_data, conn, mapping_rec['external_create_method'], record_data['id'], context=context)
                            create_id = record_data['id']
                            self.create_external_id_vals(cr, uid, record_data['id'], crid, referential_id, context=context)
                            _logger.info("Created in External Ref %s from OpenERP with external_id %s and OpenERP id %s successfully",
                                         self._name, crid, record_data['id'])
                        except Exception, err:
                            _logger.error("Failed to create in External Ref %s from OpenERP with OpenERP id %s",
                                          self._name, record_data['id'], exc_info=True)
                            raise

        return create_id, write_id

    def get_oe_user(self, cr, uid, rm_user_id, context=None):
        users_pool = self.pool.get('res.users')
        user_ids = users_pool.search(cr, uid, [('rm_user_id','=',rm_user_id)], limit=1, context=context)
        if user_ids:
            return user_ids[0]
        else:
            return False

    def _get_oe_res(self, cr, uid, external_referential_id, rm_id, oe_model_name, context=None):
        data_pool = self.pool.get('ir.model.data')
        ids = data_pool.search(cr, uid, [('external_referential_id','=',external_referential_id),
                                         ('model','=',oe_model_name),
                                         ('name','=','%s/%s' % (oe_model_name.replace('.','_'), rm_id))],
                               context=context)
        ids = [r['res_id'] for r in data_pool.read(cr, uid, ids, ['res_id'], context=context)]
        if ids:
            return ids[-1]
        else:
            return False

    def get_oe_project(self, cr, uid, external_referential_id, rm_project_id, context=None):
        return self._get_oe_res(cr, uid, external_referential_id, rm_project_id, 'project.project', context=context)

    def get_oe_task(self, cr, uid, external_referential_id, rm_task_id, context=None):
        return self._get_oe_res(cr, uid, external_referential_id, rm_task_id, 'project.task', context=context)
